import pytest

from .conftest import StrSource


@pytest.mark.asyncio
async def test_source_generator(source: StrSource) -> None:
    assert source.is_resource_allocated is False

    batches: list[list[str]] = []

    generator = source.read()

    for n in range(5):
        batch = await anext(generator)
        assert source.is_resource_allocated is True
        batches.append(batch)

    await generator.aclose()
    assert source.is_resource_allocated is False

    assert batches == [
        source.BATCHES[0],
        source.BATCHES[1],
        source.BATCHES[0],
        source.BATCHES[1],
        source.BATCHES[0],
    ]



import asyncio
import logging
from collections.abc import Sequence
from dataclasses import dataclass

from .worker import Worker, WorkerSettings

log = logging.getLogger(__name__)

DEFAULT_GROUP = "default"


@dataclass
class WorkerDef:
    worker_class: type[Worker]
    worker_settings: WorkerSettings
    worker_group: str = DEFAULT_GROUP
    concurrency: int = 1

    @property
    def worker_type(self) -> str:
        return self.worker_class.worker_type


@dataclass
class WorkerPoolSettings:
    workers: Sequence[WorkerDef]


class WorkerPool:
    settings: WorkerPoolSettings
    worker_defs: dict[str, WorkerDef]

    def __init__(
        self,
        settings: WorkerPoolSettings,
        include_groups: Sequence[str] | None = None,
        exclude_types: Sequence[str] | None = None,
        burst: bool = False,
    ) -> None:
        self.settings = settings
        self.include_groups = include_groups
        self.exclude_types = exclude_types
        self.burst = burst
        self.worker_defs = {}
        self._terminate = asyncio.Event()

    def _filter_workers(self) -> None:
        self.worker_defs = {}

        for worker_def in self.settings.workers:
            if (
                self.include_groups is not None
                and worker_def.worker_group not in self.include_groups
            ):
                log.debug(
                    f"Skipping {worker_def} because worker group not in include groups"
                )
                continue

            worker_type = worker_def.worker_type

            if self.exclude_types is not None and worker_type in self.exclude_types:
                log.debug(f"Skipping {worker_def} because worker type in exclude types")
                continue

            assert worker_type not in self.worker_defs
            self.worker_defs[worker_type] = worker_def

    async def _run_workers(self) -> None:
        pool: dict[str, list[asyncio.Task | None]] = {}
        run_count: dict[str, list[int]] = {}

        for worker_type, worker_def in self.worker_defs.items():
            pool[worker_type] = [None] * worker_def.concurrency
            run_count[worker_type] = [0] * worker_def.concurrency

        while True:
            has_done_task = False
            has_run_task = False

            for worker_type, worker_def in self.worker_defs.items():
                concurrency = len(pool[worker_type])
                for worker_num in range(concurrency):
                    task = pool[worker_type][worker_num]
                    if task is None:
                        continue

                    if task.done():
                        error = task.exception()
                        if error is not None:
                            log.error(
                                "Error in worker %s[#%d]: %s",
                                worker_type,
                                worker_num,
                                str(error),
                                exc_info=error,
                            )
                            # sentry_sdk.capture_exception(error)
                        else:
                            if not self.burst:
                                log.error(
                                    "Worker %s[#%d] was exited unexpectedly",
                                    worker_type,
                                    worker_num,
                                )
                        pool[worker_type][worker_num] = None
                        has_done_task = True
                    else:
                        has_run_task = True

            if has_done_task and not self.burst:
                await asyncio.sleep(10)

            for worker_type, worker_def in self.worker_defs.items():
                concurrency = len(pool[worker_type])
                for worker_num in range(concurrency):
                    task = pool[worker_type][worker_num]

                    if self._terminate.is_set():
                        if task is not None:
                            log.info(f"Terminating worker {worker_type}[#{worker_num}]")
                            task.cancel()
                    else:
                        if task is None:
                            no_start = (
                                self.burst and run_count[worker_type][worker_num] > 0
                            )
                            if not no_start:
                                log.info(
                                    f"Starting worker {worker_type}[#{worker_num}]"
                                )
                                kw = {}
                                if self.burst:
                                    kw["burst"] = True
                                task = asyncio.create_task(
                                    worker_def.worker_class(
                                        worker_def.worker_settings
                                    ).run(**kw)
                                )
                                pool[worker_type][worker_num] = task
                                run_count[worker_type][worker_num] += 1
                                has_run_task = True

            if self.burst and not has_run_task:
                self._terminate.set()

            all_tasks = [
                task for task in sum(list(pool.values()), []) if task is not None
            ]

            if self._terminate.is_set():
                await asyncio.gather(*all_tasks, return_exceptions=True)
                break

            await asyncio.wait(
                all_tasks + [asyncio.create_task(self._terminate.wait())],
                return_when=asyncio.FIRST_COMPLETED,
            )

    async def run(self) -> None:
        self._filter_workers()
        await self._run_workers()

    def terminate(self) -> None:
        self._terminate.set()



import asyncio
from asyncio.exceptions import CancelledError
from dataclasses import dataclass

import pytest

from aiosafeconsumer import (
    Worker,
    WorkerDef,
    WorkerPool,
    WorkerPoolSettings,
    WorkerSettings,
)


@dataclass
class AllWorkerSettings(WorkerSettings):
    counter: dict[str, int]


class BaseWorker(Worker):
    settings: AllWorkerSettings

    async def run(self, burst: bool = False) -> None:
        counter = self.settings.counter
        counter.setdefault(self.worker_type, 0)

        while True:
            counter[self.worker_type] += 1

            await asyncio.sleep(0.01)
            if burst:
                break


class Worker1(BaseWorker):
    worker_type = "worker1"


class Worker2(BaseWorker):
    worker_type = "worker2"


class Worker3(BaseWorker):
    worker_type = "worker3"


@pytest.fixture
def counter() -> dict[str, int]:
    return {}


@pytest.fixture
def worker_pool_settings(counter: dict[str, int]) -> WorkerPoolSettings:
    workers = [
        WorkerDef(
            worker_class=Worker1,
            worker_settings=AllWorkerSettings(counter=counter),
        ),
        WorkerDef(
            worker_class=Worker2,
            worker_settings=AllWorkerSettings(counter=counter),
        ),
        WorkerDef(
            worker_class=Worker3,
            worker_settings=AllWorkerSettings(counter=counter),
        ),
    ]
    return WorkerPoolSettings(workers=workers)


@pytest.mark.asyncio
async def test_worker_pool_run(
    counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    pool = WorkerPool(worker_pool_settings)

    assert counter == {}

    task = asyncio.create_task(pool.run())
    await asyncio.sleep(0.1)
    task.cancel()

    with pytest.raises(CancelledError):
        await task

    assert counter["worker1"] > 1
    assert counter["worker2"] > 1
    assert counter["worker3"] > 1


@pytest.mark.asyncio
async def test_worker_pool_run_burst(
    counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    assert counter == {}

    pool = WorkerPool(worker_pool_settings, burst=True)
    await pool.run()

    assert counter == {
        "worker1": 1,
        "worker2": 1,
        "worker3": 1,
    }



import asyncio
import dataclasses
import logging
import random
from collections.abc import AsyncGenerator, Callable
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Generic

if not TYPE_CHECKING:
    from aiokafka import AIOKafkaConsumer, ConsumerRebalanceListener, TopicPartition
    from aiokafka.errors import KafkaConnectionError, KafkaError
else:
    from .kafka_stubs import (
        AIOKafkaConsumer,
        ConsumerRebalanceListener,
        KafkaConnectionError,
        KafkaError,
        TopicPartition,
    )

from .source import DataSource, DataSourceSettings
from .types import DataType

log = logging.getLogger(__name__)

CONSUMER_DEFAULTS = {
    "auto_offset_reset": "latest",
    "max_partition_fetch_bytes": 256 * 1024,
    "fetch_min_bytes": 32 * 1024,
    "fetch_max_wait_ms": 2500,
    "session_timeout_ms": 10000,
    "max_poll_interval_ms": 300000,
}


class InformativeRebalanceListener(ConsumerRebalanceListener):
    revoked: set[TopicPartition]
    assigned: set[TopicPartition]

    def __init__(self) -> None:
        self.revoked = set()
        self.assigned = set()

    async def on_partitions_revoked(self, revoked: list[TopicPartition]) -> None:
        self.revoked = set(revoked)

    async def on_partitions_assigned(self, assigned: list[TopicPartition]) -> None:
        self.assigned = set(assigned)


@dataclasses.dataclass
class KafkaSourceSettings(Generic[DataType], DataSourceSettings):
    topics: list[str]
    bootstrap_servers: list[str]
    value_deserializer: Callable[[bytes], DataType]
    group_id: str | None = None
    enable_auto_commit: bool = False
    max_read_partitions: int | None = None
    getmany_max_records: int | None = None
    getmany_timeout: timedelta = timedelta(seconds=10)
    max_consumer_errors: int = 100
    sleep_on_consumer_error: timedelta = timedelta(seconds=5)
    kwargs: dict[str, Any] = dataclasses.field(default_factory=dict)


class KafkaSource(Generic[DataType], DataSource[DataType]):
    settings: KafkaSourceSettings[DataType]

    def __init__(self, settings: KafkaSourceSettings[DataType]):
        super().__init__(settings)
        self.commit_event = asyncio.Event()

    def __str__(self) -> str:
        return self.__class__.__name__

    def get_topics(self) -> list[str]:
        return self.settings.topics

    def get_bootstrap_servers(self) -> list[str]:
        servers = list(self.settings.bootstrap_servers)
        random.shuffle(servers)
        return servers

    def get_default_group_id(self) -> str:
        return self.settings.group_id or self.__class__.__name__

    def create_kafka_consumer(self) -> AIOKafkaConsumer:
        kwargs = {
            **CONSUMER_DEFAULTS,
            "bootstrap_servers": self.get_bootstrap_servers(),
            "enable_auto_commit": self.settings.enable_auto_commit,
            "group_id": self.get_default_group_id(),
            "value_deserializer": self.settings.value_deserializer,
            **self.settings.kwargs,
        }
        return AIOKafkaConsumer(**kwargs)

    def create_listener(self) -> InformativeRebalanceListener:
        return InformativeRebalanceListener()

    async def read(self) -> AsyncGenerator[list[DataType], None]:
        try:
            consumer = None
            listener = self.create_listener()
            consumer_errors_count = 0

            while True:

                if (
                    consumer_errors_count > self.settings.max_consumer_errors
                    and consumer is not None
                ):
                    await consumer.stop()
                    consumer = None
                    consumer_errors_count = 0
                    log.error("To many consumer errors, restarting consumer")

                while consumer is None:
                    consumer = self.create_kafka_consumer()
                    try:
                        consumer.subscribe(self.get_topics(), listener=listener)
                        await consumer.start()
                    except KafkaConnectionError as error:
                        log.error(f"Kafka error: {error}")
                        consumer = None
                        await asyncio.sleep(
                            self.settings.sleep_on_consumer_error.total_seconds()
                        )

                listener.revoked = set()
                consumed_data = await consumer.getmany(
                    timeout_ms=self.settings.getmany_timeout.total_seconds() * 1000,
                    max_records=self.settings.getmany_max_records,
                )
                consumed_count = sum(len(x) for x in consumed_data.values())
                if not consumed_count:
                    continue

                skipped_tps = set()
                consumed_tps = set()

                for tp, messages in consumed_data.items():
                    if tp in listener.revoked:
                        continue

                    if (
                        self.settings.max_read_partitions
                        and len(consumed_tps) >= self.settings.max_read_partitions
                    ):
                        skipped_tps.add(tp)
                        continue

                    consumed_tps.add(tp)

                    yield [msg.value for msg in messages]

                if listener.revoked:
                    count = sum(
                        len(consumed_data.get(tp, [])) for tp in listener.revoked
                    )
                    log.debug(
                        f"Will not process {count} messages in {len(listener.revoked)}"
                        " partitions because partitions assignment is revoked"
                    )
                if skipped_tps:
                    count = sum(len(consumed_data.get(tp, [])) for tp in skipped_tps)
                    log.debug(
                        f"Will not process {count} messages in {len(skipped_tps)}"
                        " partitions because max read partitions limit achieved"
                    )

                if not self.settings.enable_auto_commit:
                    log.debug("Waiting for commit event")
                    await self.commit_event.wait()

                    no_commit_tps = listener.revoked | skipped_tps
                    for tp, messages in consumed_data.items():
                        if tp not in no_commit_tps:
                            new_offset = messages[-1].offset + 1
                            log.debug(f"Commiting {tp} to {new_offset}")
                            try:
                                await consumer.commit({tp: new_offset})
                            except KafkaError as error:
                                self.log_error(f"Kafka error: {error}")
                                consumer_errors_count += 1

        finally:
            if consumer is not None:
                await consumer.stop()

    async def commit(self, batch: list[DataType]) -> None:
        self.commit_event.set()



from contextvars import ContextVar

VAR_PREFIX = "aiosafeconsumer."

worker_type_context: ContextVar[str | None] = ContextVar(
    f"{VAR_PREFIX}worker_type",
    default=None,
)
worker_id_context: ContextVar[str | None] = ContextVar(
    f"{VAR_PREFIX}worker_id",
    default=None,
)



class TopicPartition:
    pass


class ConsumerRebalanceListener:
    async def on_partitions_revoked(self, revoked: list[TopicPartition]) -> None:
        pass

    async def on_partitions_assigned(self, assigned: list[TopicPartition]) -> None:
        pass


class AIOKafkaConsumer:
    def subscribe(
        self, topics: list[str], listener: ConsumerRebalanceListener | None = None
    ) -> None:
        pass

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass


class KafkaError(Exception):
    pass


class KafkaConnectionError(KafkaError):
    pass



import logging
from collections.abc import Callable
from dataclasses import dataclass
from datetime import timedelta
from typing import Generic

from redis.asyncio import Redis

from ..types import DataType
from .base import DataWriter, DataWriterSettings
from .types import EnumerateIDsRecord, EventType, ObjectID, Version

log = logging.getLogger(__name__)


@dataclass
class RedisWriterSettings(Generic[DataType], DataWriterSettings[DataType]):
    redis: Callable[[], Redis]
    version_serializer: Callable[[Version], bytes]
    version_deserializer: Callable[[bytes], Version]
    record_serializer: Callable[[DataType], bytes]
    key_prefix: str
    versions_key: str
    versions_expire: timedelta = timedelta(hours=24)
    record_expire: timedelta = timedelta(hours=30)


class RedisWriter(Generic[DataType], DataWriter[DataType]):
    settings: RedisWriterSettings

    def _obj_key(self, obj_id: ObjectID) -> bytes:
        return f"{self.settings.key_prefix}{obj_id}".encode()

    def _obj_version_key(self, obj_id: ObjectID) -> bytes:
        return str(obj_id).encode()

    def _obj_version_key_to_obj_key(self, val: bytes) -> bytes:
        key_prefix = self.settings.key_prefix.encode()
        return key_prefix + val

    def _versions_key(self) -> bytes:
        return self.settings.versions_key.encode()

    async def process(self, batch: list[DataType]) -> None:
        settings = self.settings
        redis = settings.redis()

        enum_records: dict[Version, list[EnumerateIDsRecord]] = {}
        upsert_records: dict[ObjectID, DataType] = {}
        upsert_versions: dict[ObjectID, Version] = {}

        versions = await redis.hgetall(self._versions_key())
        log.debug(f"Loaded {len(versions)} object versions from Redis")

        for record in batch:
            event_type = settings.event_type_getter(record)
            rec_ver = settings.version_getter(record)

            if event_type == EventType.ENUMERATE:
                enum_rec = settings.enum_getter(record)
                enum_records.setdefault(rec_ver, [])
                enum_records[rec_ver].append(enum_rec)
            else:
                obj_id = settings.id_getter(record)
                cur_ver = upsert_versions.get(obj_id)
                if cur_ver is None:
                    cur_ver = settings.version_deserializer(
                        versions.get(self._obj_version_key(obj_id), b"0")
                    )

                if rec_ver > cur_ver:  # type: ignore
                    upsert_records[obj_id] = record
                    upsert_versions[obj_id] = rec_ver

        update_records: dict[bytes, bytes] = {}
        update_versions: dict[bytes, bytes] = {}
        del_versions: list[bytes] = []
        del_records: list[bytes] = []

        for obj_id, record in upsert_records.items():
            event_type = settings.event_type_getter(record)
            ver_key = self._obj_version_key(obj_id)
            key = self._obj_key(obj_id)
            if event_type == EventType.DELETE:
                del_versions.append(ver_key)
                del_records.append(key)
                continue
            value = settings.record_serializer(record)
            update_records[key] = value
            ver_value = settings.version_serializer(upsert_versions[obj_id])
            update_versions[ver_key] = ver_value

        if enum_records:
            # TODO: implement sessions with chunks of IDs.
            # Now just use last non chunked record.
            use_enum_rec: EnumerateIDsRecord | None = None
            for rec_ver in reversed(sorted(enum_records.keys())):
                for rec in enum_records[rec_ver]:
                    if rec.chunk is None:
                        use_enum_rec = rec
                        break
                if use_enum_rec is not None:
                    break

            if use_enum_rec is not None:
                cur_ids: set[bytes] = set(
                    list(versions.keys())
                    + [
                        self._obj_version_key(obj_id)
                        for obj_id in upsert_versions.keys()
                    ]
                )
                enum_ids: set[bytes] = set(
                    self._obj_version_key(obj_id) for obj_id in use_enum_rec.ids
                )
                ids_to_delete: set[bytes] = cur_ids - enum_ids
                log.debug(
                    f"Accept enumerate message with {len(enum_ids)} object IDs,"
                    f" {len(ids_to_delete)} records will be deleted"
                )
                del_records.extend(
                    [self._obj_version_key_to_obj_key(v) for v in ids_to_delete]
                )
                del_versions.extend(ids_to_delete)

        async with redis.pipeline(transaction=True) as pipe:
            if update_records:
                pipe.mset(update_records)  # type: ignore
            for key in update_records.keys():
                pipe.expire(key, settings.record_expire)
            if del_records:
                pipe.delete(*del_records)
            if del_versions:
                pipe.hdel(self._versions_key(), *del_versions)
            if update_versions:
                pipe.hset(self._versions_key(), mapping=update_versions)  # type: ignore
            pipe.expire(self._versions_key(), settings.versions_expire)
            await pipe.execute()

        log.debug(
            f"{len(update_records)} records was added/updated and {len(del_records)}"
            " records was deleted"
        )



from abc import ABC
from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic

from ..processor import DataProcessor, DataProcessorSettings
from ..types import DataType
from .types import EnumerateIDsRecord, EventType, ObjectID, Version


@dataclass
class DataWriterSettings(Generic[DataType], DataProcessorSettings):
    version_getter: Callable[[DataType], Version]
    event_type_getter: Callable[[DataType], EventType]
    id_getter: Callable[[DataType], ObjectID]
    enum_getter: Callable[[DataType], EnumerateIDsRecord]


class DataWriter(Generic[DataType], DataProcessor[DataType], ABC):
    settings: DataWriterSettings



import json
from dataclasses import dataclass
from datetime import timedelta
from typing import NamedTuple, cast
from unittest import mock

import pytest
from redis.asyncio import Redis
from redis.asyncio.client import Pipeline

from aiosafeconsumer.datasync import EnumerateIDsRecord, EventType, ObjectID, Version
from aiosafeconsumer.datasync.redis import RedisWriter, RedisWriterSettings


class ItemRecord(NamedTuple):
    ev_type: EventType
    version: int
    id: int
    data: str


class ItemDeleteRecord(NamedTuple):
    ev_type: EventType
    version: int
    id: int


class ItemEnumerateRecord(NamedTuple):
    ev_type: EventType
    version: int
    ids: list[int]


Item = ItemRecord | ItemDeleteRecord | ItemEnumerateRecord


@dataclass
class Settings(RedisWriterSettings[Item]):
    pass


class Writer(RedisWriter[Item]):
    settings: Settings


@pytest.fixture
def redis_mock() -> Redis:
    versions: dict[bytes, bytes] = {}
    db: dict[bytes, bytes] = {}
    expire_db: dict[bytes, timedelta] = {}

    async def hgetall(key: bytes) -> dict[bytes, bytes]:
        if key == b"items":
            return versions.copy()
        return {}

    def hdel(name: bytes, *keys: bytes) -> None:
        if name == b"items":
            for k in keys:
                versions.pop(k, None)

    def hset(name: bytes, mapping: dict[bytes, bytes]) -> None:
        if name == b"items":
            versions.update(mapping)

    def mset(mapping: dict[bytes, bytes]) -> None:
        db.update(mapping)

    def delete(*keys: bytes) -> None:
        for key in keys:
            db.pop(key, None)

    def expire(key: bytes, ex: timedelta) -> None:
        expire_db[key] = ex

    redis = mock.MagicMock(spec=Redis)
    redis.hgetall.side_effect = hgetall

    pipe = mock.MagicMock(spec=Pipeline)
    redis.pipeline.return_value.__aenter__.return_value = pipe

    pipe.delete.side_effect = delete
    pipe.expire.side_effect = expire
    pipe.hdel.side_effect = hdel
    pipe.hset.side_effect = hset
    pipe.mset.side_effect = mset

    redis._test_versions = versions
    redis._test_db = db
    redis._test_expire_db = expire_db

    return redis


@pytest.fixture
def settings(redis_mock: Redis) -> Settings:
    ENCODING = "utf-8"

    def get_redis() -> Redis:
        return redis_mock

    def version_getter(item: Item) -> Version:
        return item.version

    def record_serializer(item: Item) -> bytes:
        return json.dumps(item._asdict()).encode(ENCODING)

    def event_type_getter(item: Item) -> EventType:
        return item.ev_type

    def id_getter(item: Item) -> ObjectID:
        assert isinstance(item, ItemRecord) or isinstance(item, ItemDeleteRecord)
        return item.id

    def enum_getter(item: Item) -> EnumerateIDsRecord:
        assert isinstance(item, ItemEnumerateRecord)
        return EnumerateIDsRecord(ids=cast(list[ObjectID], item.ids))

    def version_serializer(ver: Version) -> bytes:
        return str(ver).encode()

    def version_deserializer(val: bytes) -> Version:
        return int(val.decode())

    settings = Settings(
        redis=get_redis,
        version_getter=version_getter,
        version_serializer=version_serializer,
        version_deserializer=version_deserializer,
        record_serializer=record_serializer,
        key_prefix="item:",
        versions_key="items",
        event_type_getter=event_type_getter,
        id_getter=id_getter,
        enum_getter=enum_getter,
    )
    return settings


@pytest.mark.asyncio
async def test_redis_writer_initial_empty(
    redis_mock: Redis, settings: Settings
) -> None:
    items: list[Item] = [
        ItemRecord(ev_type=EventType.CREATE, version=1, id=1, data="1v1"),
        ItemRecord(ev_type=EventType.UPDATE, version=2, id=2, data="2v2"),
        ItemRecord(ev_type=EventType.REFRESH, version=1, id=3, data="3v1"),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"1",
        b"2": b"2",
        b"3": b"1",
    }
    assert redis_mock._test_db == {  # type: ignore
        b"item:1": b'{"ev_type": "create", "version": 1, "id": 1, "data": "1v1"}',
        b"item:2": b'{"ev_type": "update", "version": 2, "id": 2, "data": "2v2"}',
        b"item:3": b'{"ev_type": "refresh", "version": 1, "id": 3, "data": "3v1"}',
    }
    assert redis_mock._test_expire_db == {  # type: ignore
        b"item:1": timedelta(hours=30),
        b"item:2": timedelta(hours=30),
        b"item:3": timedelta(hours=30),
        b"items": timedelta(hours=24),
    }


@pytest.mark.asyncio
async def test_redis_writer_upsert(redis_mock: Redis, settings: Settings) -> None:
    redis_mock._test_versions.update(  # type: ignore
        {
            b"1": b"0",
            b"2": b"2",
            b"3": b"1",
            b"4": b"0",
        },
    )
    redis_mock._test_db.update(  # type:ignore
        {
            b"item:1": b"",
            b"item:2": b"",
            b"item:3": b"",
            b"item:4": b"",
        },
    )

    items: list[Item] = [
        ItemRecord(ev_type=EventType.CREATE, version=1, id=1, data="1v1"),
        ItemRecord(ev_type=EventType.UPDATE, version=2, id=2, data="2v2"),
        ItemRecord(ev_type=EventType.REFRESH, version=2, id=3, data="3v2"),
        ItemRecord(ev_type=EventType.DELETE, version=1, id=4, data=""),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"1",
        b"2": b"2",
        b"3": b"2",
    }
    assert redis_mock._test_db == {  # type: ignore
        b"item:1": b'{"ev_type": "create", "version": 1, "id": 1, "data": "1v1"}',
        b"item:2": b"",
        b"item:3": b'{"ev_type": "refresh", "version": 2, "id": 3, "data": "3v2"}',
    }
    assert redis_mock._test_expire_db == {  # type: ignore
        b"item:1": timedelta(hours=30),
        b"item:3": timedelta(hours=30),
        b"items": timedelta(hours=24),
    }


@pytest.mark.asyncio
async def test_redis_writer_enumerate(redis_mock: Redis, settings: Settings) -> None:
    redis_mock._test_versions.update(  # type: ignore
        {
            b"1": b"0",
            b"2": b"2",
            b"3": b"1",
            b"4": b"0",
        },
    )
    redis_mock._test_db.update(  # type:ignore
        {
            b"item:1": b"",
            b"item:2": b"",
            b"item:3": b"",
            b"item:4": b"",
        },
    )

    items: list[Item] = [
        ItemEnumerateRecord(ev_type=EventType.ENUMERATE, version=1, ids=[1, 3]),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"0",
        b"3": b"1",
    }
    assert redis_mock._test_db == {  # type: ignore
        b"item:1": b"",
        b"item:3": b"",
    }
    assert redis_mock._test_expire_db == {  # type: ignore
        b"items": timedelta(hours=24),
    }



from datetime import datetime
from enum import Enum
from typing import NamedTuple, TypeAlias


class EventType(str, Enum):
    CREATE = "create"  # object was createed
    UPDATE = "update"  # object was updated
    DELETE = "delete"  # object was deleted
    REFRESH = "refresh"  # nothing happened to the object, just refresh data
    ENUMERATE = "enumerate"  # enumerate IDs of existing objects
    EOS = "eos"  # end of stream


Version: TypeAlias = int | str | datetime
ObjectID: TypeAlias = int | str


class EnumerateIDsChunk(NamedTuple):
    chunk_index: int
    total_chunks: int
    session: str


class EnumerateIDsRecord(NamedTuple):
    ids: list[ObjectID]
    chunk: EnumerateIDsChunk | None = None



import argparse
import asyncio
import functools
import logging
import multiprocessing
import signal
import sys
import threading
import time
from datetime import datetime
from importlib import import_module
from typing import Any

from croniter import croniter

from .workerpool import WorkerPool, WorkerPoolSettings

log = logging.getLogger(__name__)


def _parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="aiosafeconsumer worker")
    parser.add_argument(
        "--concurrency",
        "-c",
        metavar="NUMBER",
        type=int,
        default=None,
        help="Override workers concurrency",
    )
    parser.add_argument(
        "--processes",
        "-p",
        metavar="NUMBER",
        type=int,
        default=1,
        help="Run NUMBER of child processes",
    )
    parser.add_argument(
        "--task",
        "-t",
        metavar="TASK_TYPE",
        dest="tasks",
        action="append",
        help="Run only specified tasks. To see available tasks use --list-tasks",
    )
    parser.add_argument(
        "--exclude-task",
        "-e",
        metavar="TASK_TYPE",
        dest="exclude_types",
        action="append",
        help="Exclude tasks",
    )
    parser.add_argument(
        "--list-tasks",
        dest="list_tasks",
        action="store_true",
        help="List available tasks",
    )
    parser.add_argument(
        "--group",
        "-g",
        metavar="TASK_GROUP",
        dest="include_groups",
        action="append",
        help=(
            "Run only tasks in specified groups. To see available groups use"
            " --list-groups"
        ),
    )
    parser.add_argument(
        "--list-groups",
        dest="list_groups",
        action="store_true",
        help="List available groups",
    )
    parser.add_argument(
        "--burst",
        dest="burst",
        action="store_true",
        default=False,
        help="Run one iteration and exit",
    )
    parser.add_argument(
        "--restart-schedule",
        metavar="CRONTAB",
        dest="restart_schedule",
        default=None,
        help="Restart processes using cron-like schedule",
    )
    parser.add_argument(
        "--pool-class",
        metavar="CLASS",
        dest="pool_class",
        default=None,
        help="WorkerPool class",
    )
    parser.add_argument(
        "--pool-settings",
        metavar="CLASS",
        dest="pool_settings",
        default=None,
        help="WorkerPoolSettings class",
    )
    return parser


def _import_name(pathname: str) -> Any:
    module_name, name = pathname.rsplit(".", 1)
    module = import_module(module_name)
    value = getattr(module, name)
    return value


def _create_pool(
    pool_class: type[WorkerPool],
    pool_settings: WorkerPoolSettings,
    args: argparse.Namespace,
) -> WorkerPool:
    pool = pool_class(
        pool_settings,
        include_groups=args.include_groups,
        exclude_types=args.exclude_types,
        burst=args.burst,
    )
    return pool


def _run_sp(
    pool_class: type[WorkerPool],
    pool_settings: WorkerPoolSettings,
    args: argparse.Namespace,
    process_id: int | None = None,
) -> None:
    def handle_signal(sig_number: int, pool: WorkerPool) -> None:
        sig = signal.Signals(sig_number)
        log.info(
            f"terminating worker process #{(process_id or 0) + 1} by {sig.name}...",
        )
        pool.terminate()

    async def shutdown(loop: asyncio.AbstractEventLoop, pool: WorkerPool) -> None:
        log.info("waiting for all tasks complete...")
        pool.terminate()
        tasks = [
            task for task in asyncio.all_tasks() if task is not asyncio.current_task()
        ]
        await asyncio.gather(*tasks, return_exceptions=True)

    # logging.config.dictConfig(config.logging)
    pool = _create_pool(pool_class, pool_settings, args)
    loop = asyncio.new_event_loop()
    loop.add_signal_handler(
        signal.SIGINT, functools.partial(handle_signal, signal.SIGINT, pool)
    )
    loop.add_signal_handler(
        signal.SIGTERM, functools.partial(handle_signal, signal.SIGTERM, pool)
    )

    try:
        loop.run_until_complete(pool.run())
    except KeyboardInterrupt:
        pass
    finally:
        loop.run_until_complete(shutdown(loop, pool))
    loop.close()


def _run_mp(
    pool_class: type[WorkerPool],
    pool_settings: WorkerPoolSettings,
    args: argparse.Namespace,
) -> None:
    def create_process(
        args: argparse.Namespace, process_id: int
    ) -> multiprocessing.Process:
        proc = multiprocessing.Process(
            target=_run_sp,
            args=[args, process_id],
            name=f"process-{process_id + 1}",
            daemon=True,
        )
        return proc

    def handle_signal(sig_number: int, stack: Any) -> None:
        sig = signal.Signals(sig_number)
        log.info(f"Terminating master process by {sig.name}")
        terminate.set()

    # logging.config.dictConfig(config.logging)

    if args.restart_schedule:
        restart_iter = croniter(args.restart_schedule)
        restart_at = restart_iter.get_next()
        restart_at_str = datetime.utcfromtimestamp(restart_at).isoformat()
        log.debug("Will restart at {restart_at_str} UTC")
    else:
        restart_iter = None
        restart_at = None

    terminate = threading.Event()
    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)
    children = [
        create_process(args, process_id) for process_id in range(args.processes)
    ]

    for proc in children:
        proc.start()

    try:
        while not terminate.is_set():
            is_restart = restart_at and restart_at <= time.time()
            if is_restart:
                for process_id, proc in enumerate(children):
                    log.info(f"Restarting process #{process_id + 1} by schedule")
                    proc.terminate()
                for proc in children:
                    proc.join()

            for process_id, proc in enumerate(children):
                if not proc.is_alive():
                    if not is_restart:
                        log.info(f"Process %{process_id + 1} was died, restarting")
                    proc.close()
                    new_proc = create_process(args, process_id)
                    new_proc.start()
                    children[process_id] = new_proc

            if is_restart and restart_iter:
                while restart_at <= time.time():
                    restart_at = restart_iter.get_next()
                restart_at_str = datetime.utcfromtimestamp(restart_at).isoformat()
                log.debug(f"Will restart at {restart_at_str} UTC")
            terminate.wait(1)
    finally:
        log.info("Terminating all child processes")
        for proc in children:
            proc.terminate()
        for proc in children:
            proc.join()
            proc.close()


def _list_tasks(pool_settings: WorkerPoolSettings) -> None:
    tasks: set[str] = set()
    for worker_def in pool_settings.workers:
        tasks.add(worker_def.worker_type)
    for task in sorted(tasks):
        print(task)


def _list_groups(pool_settings: WorkerPoolSettings) -> None:
    groups: set[str] = set()
    for worker_def in pool_settings.workers:
        groups.add(worker_def.worker_group)
    for group in sorted(groups):
        print(group)


def run(
    argv: list[str] | None = None,
    pool_class: type[WorkerPool] | None = None,
    pool_settings: WorkerPoolSettings | None = None,
) -> None:
    parser = _parser()
    args = parser.parse_args(argv and argv[1:] or [])

    if args.pool_class:
        value = _import_name(args.pool_class)
        assert issubclass(value, WorkerPool)
        pool_class = value

    if args.pool_settings:
        value = _import_name(args.pool_settings)
        if callable(value):
            value = value()
        assert isinstance(value, WorkerPoolSettings)
        pool_settings = value

    assert (
        pool_class
    ), "The pool_class is required. Pass it as function argument or cli argument"
    assert (
        pool_settings
    ), "The pool_settings is required. Pass it as function argument or cli argument"

    if args.list_tasks:
        _list_tasks(pool_settings)
    elif args.list_groups:
        _list_groups(pool_settings)
    else:
        if args.processes > 1 or args.restart_schedule:
            _run_mp(pool_class, pool_settings, args)
        else:
            _run_sp(pool_class, pool_settings, args)


if __name__ == "__main__":
    run(argv=sys.argv, pool_class=WorkerPool)
